@@ -1,19 +1,28 @@
 from __future__ import unicode_literals
 
+import os
 import re
 import cgi
 import json
 import urllib
+import codecs
+
+import sys
+PY3 = (sys.version_info[0] == 3)
+if PY3:
+    unicode = str
+    basestring = str
 
 # need to add group for literals
 # switch to using word boundary for params section
 node_re = re.compile(r'({'
                      r'(?P<closing>\/)?'
-                     r'(?P<symbol>[\~\#\?\@\:\<\>\+\^])?'
+                     r'(?:(?P<symbol>[\~\#\?\@\:\<\>\+\^\%])\s*)?'
                      r'(?P<refpath>[a-zA-Z0-9_\$\.]+|"[^"]+")'
                      r'(?:\:(?P<contpath>[a-zA-Z0-9\$\.]+))?'
                      r'(?P<filters>\|[a-z]+)*?'
-                     r'(?P<params> \w+\=(("[^"]*?")|([\w\.]+)))*?'
+                     r'(?P<params>(?:\s+\w+\=(("[^"]*?")|([\w\.]+)))*)?'
+                     r'\s*'
                      r'(?P<selfclosing>\/)?'
                      r'\})',
                      flags=re.MULTILINE)
@@ -21,65 +30,90 @@ node_re = re.compile(r'({'
 key_re_str = '[a-zA-Z_$][0-9a-zA-Z_$]*'
 key_re = re.compile(key_re_str)
 path_re = re.compile('(' + key_re_str + ')?(\.' + key_re_str + ')+')
-
-
-#comment_re = ''  # TODO
-def strip_comments(text):
-    return re.sub(r'\{!.+?!\}', '', text, flags=re.DOTALL)
+comment_re = re.compile(r'(\{!.+?!\})', flags=re.DOTALL)
 
 
 def get_path_or_key(pork):
     if pork == '.':
-        pk = ('path', True, [])
+        pk = ['path', True, []]
     elif path_re.match(pork):
         f_local = pork.startswith('.')
         if f_local:
             pork = pork[1:]
-        pk = ('path', f_local, pork.split('.'))
+        pk = ['path', f_local, pork.split('.')]
     elif key_re.match(pork):
-        pk = ('key', pork)
+        pk = ['key', pork]
     else:
         raise ValueError('expected a path or key, not %r' % pork)
     return pk
 
 
-def params_to_kv(params_str):
-    ret = []
-    new_k, v = None, None
-    k, _, tail = params_str.partition('=')
-    while tail:
-        tmp, _, tail = tail.partition('=')
-        if not tail:
-            v = tmp
-        else:
-            v, new_k = tmp.split()
-        ret.append((k.strip(), v.strip()))
-        k = new_k
-    return ret
+def split_leading(text):
+    leading_stripped = text.lstrip()
+    leading_ws = text[:len(text) - len(leading_stripped)]
+    return leading_ws, leading_stripped
 
 
-def wrap_params(param_kv):
-    ret = []
-    for k, v in param_kv:
-        if v.startswith('"') and v.endswith('"'):
-            v = v[1:-1]
-            v_tuple = ('literal', v)
-        else:
-            v_tuple = get_path_or_key(v)
-        ret.append(('param', ('literal', k), v_tuple))
-    return ret
+class Token(object):
+    def __init__(self, text):
+        self.text = text
+
+    def get_line_count(self):
+        # returns 0 if there's only one line, because the
+        # token hasn't increased the number of lines.
+        count = len(self.text.splitlines()) - 1
+        if self.text[-1] in ('\n', '\r'):
+            count += 1
+        return count
+
+    def __repr__(self):
+        cn = self.__class__.__name__
+        disp = self.text
+        if len(disp) > 20:
+            disp = disp[:17] + '...'
+        return '%s(%r)' % (cn, disp)
+
+
+class CommentToken(Token):
+    def to_dust_ast(self):
+        return [['comment', self.text]]
+
+
+class BufferToken(Token):
+    def to_dust_ast(self):
+        # It is hard to simulate the PEG parsing in this case,
+        # especially while supporting universal newlines.
+        if not self.text:
+            return []
+        rev = []
+        remaining_lines = self.text.splitlines()
+        if self.text[-1] in ('\n', '\r'):
+            # kind of a bug in splitlines if you ask me.
+            remaining_lines.append('')
+        while remaining_lines:
+            line = remaining_lines.pop()
+            leading_ws, lstripped = split_leading(line)
+            if remaining_lines:
+                if lstripped:
+                    rev.append(['buffer', lstripped])
+                rev.append(['format', '\n', leading_ws])
+            else:
+                if line:
+                    rev.append(['buffer', line])
+        ret = list(reversed(rev))
+        return ret
 
 
 ALL_ATTRS = ('closing', 'symbol', 'refpath', 'contpath',
              'filters', 'params', 'selfclosing')
 
 
-class Tag(object):
+class Tag(Token):
     req_attrs = ()
     ill_attrs = ()
 
-    def __init__(self, **kw):
-        self.match_str = kw.pop('match_str')
+    def __init__(self, text, **kw):
+        super(Tag, self).__init__(text)
         self._attr_dict = kw
         self.set_attrs(kw)
 
@@ -90,6 +124,13 @@ class Tag(object):
         except AttributeError:
             return []
 
+    @property
+    def name(self):
+        try:
+            return self.refpath.strip().lstrip('.')
+        except (AttributeError, TypeError):
+            return None
+
     def set_attrs(self, attr_dict, raise_exc=True):
         cn = self.__class__.__name__
         all_attrs = getattr(self, 'all_attrs', ())
@@ -115,15 +156,11 @@ class Tag(object):
             setattr(self, attr, attr_dict.get(attr, ''))
         return True
 
-    def __repr__(self):
-        cn = self.__class__.__name__
-        return '%s(%r)' % (cn, self.match_str)
-
     @classmethod
     def from_match(cls, match):
-        kw = dict([(k, v) for k, v in match.groupdict().items()
-                  if v is not None])
-        kw['match_str'] = match.group(0)
+        kw = dict([(k, v.strip()) for k, v in match.groupdict().items()
+                  if v is not None and v.strip()])
+        kw['text'] = match.group(0)
         obj = cls(**kw)
         obj.orig_match = match
         return obj
@@ -178,30 +215,61 @@ class PartialTag(Tag):
         contpath = self.contpath
         if contpath:
             context.append(get_path_or_key(contpath))
-        subtokens = self.subtokens
-        if subtokens and isinstance(subtokens[0], BufferToken):
-            body = ['literal', subtokens[0].text]
-        else:
-            body = ['body']
-            for b in self.subtokens:
-                body.extend(b.to_dust_ast())
+        inline_body = inline_to_dust_ast(self.subtokens)
         return [['partial',
-                 body,
+                 inline_body,
                  context]]
 
 
 def parse_inline(source):
     if not source:
-        raise ValueError('empty inline token')
-    orig_source = source
+        raise ParseError('empty inline token')
     if source.startswith('"') and source.endswith('"'):
         source = source[1:-1]
     if not source:
-        return BufferToken()
+        return [BufferToken()]
     tokens = tokenize(source, inline=True)
     return tokens
 
 
+def inline_to_dust_ast(tokens):
+    if tokens and isinstance(tokens[0], BufferToken):
+        body = ['literal', tokens[0].text]
+    else:
+        body = ['body']
+        for b in tokens:
+            body.extend(b.to_dust_ast())
+    return body
+
+
+def params_to_kv(params_str):
+    ret = []
+    new_k, v = None, None
+    p_str = params_str.strip()
+    k, _, tail = p_str.partition('=')
+    while tail:
+        tmp, _, tail = tail.partition('=')
+        tail = tail.strip()
+        if not tail:
+            v = tmp
+        else:
+            v, new_k = tmp.split()
+        ret.append((k.strip(), v.strip()))
+        k = new_k
+    return ret
+
+
+def params_to_dust_ast(param_kv):
+    ret = []
+    for k, v in param_kv:
+        try:
+            v_body = get_path_or_key(v)
+        except ValueError:
+            v_body = inline_to_dust_ast(parse_inline(v))
+        ret.append(['param', ['literal', k], v_body])
+    return ret
+
+
 def get_tag(match, inline=False):
     groups = match.groupdict()
     symbol = groups['symbol']
@@ -220,94 +288,61 @@ def get_tag(match, inline=False):
     elif symbol == '>':
         tag_type = PartialTag
     else:
-        raise ValueError('invalid tag')
+        raise ParseError('invalid tag')
     if inline and tag_type not in (ReferenceTag, SpecialTag):
-        raise ValueError('invalid inline tag')
+        raise ParseError('invalid inline tag')
     return tag_type.from_match(match)
 
 
-class WhitespaceToken(object):
-    def __init__(self, ws=''):
-        self.ws = ws
-
-    def __repr__(self):
-        disp = self.ws
-        if len(disp) > 13:
-            disp = disp[:10] + '...'
-        return u'WhitespaceToken(%r)' % disp
-
-
-def split_leading(text):
-    leading_stripped = text.lstrip()
-    leading_ws = text[:len(text) - len(leading_stripped)]
-    return leading_ws, leading_stripped
-
-
-class BufferToken(object):
-    def __init__(self, text=''):
-        self.text = text
-
-    def __repr__(self):
-        disp = self.text
-        if len(self.text) > 30:
-            disp = disp[:27] + '...'
-        return u'BufferToken(%r)' % disp
-
-    def to_dust_ast(self):
-        # It is hard to simulate the PEG parsing in this case,
-        # especially while supporting universal newlines.
-        if not self.text:
-            return []
-        rev = []
-        remaining_lines = self.text.splitlines()
-        if self.text[-1] in ('\n', '\r'):
-            # kind of a bug in splitlines if you ask me.
-            remaining_lines.append('')
-        while remaining_lines:
-            line = remaining_lines.pop()
-            leading_ws, lstripped = split_leading(line)
-            if remaining_lines:
-                if lstripped:
-                    rev.append(['buffer', lstripped])
-                rev.append(['format', '\n', leading_ws])
-            else:
-                if line:
-                    rev.append(['buffer', line])
-        ret = list(reversed(rev))
-        return ret
-
-
 def tokenize(source, inline=False):
-    # TODO: line/column numbers
-    # removing comments
-    source = strip_comments(source)
     tokens = []
-    prev_end = 0
-    start = None
-    end = None
-    for match in node_re.finditer(source):
-        start, end = match.start(1), match.end(1)
-        if prev_end < start:
-            tokens.append(BufferToken(source[prev_end:start]))
-        prev_end = end
-        tag = get_tag(match, inline)
-        tokens.append(tag)
-    tail = source[prev_end:]
-    if tail:
-        tokens.append(BufferToken(tail))
+    com_nocom = comment_re.split(source)
+    line_counts = [1]
+    def _add_token(t):
+        # i wish i had nonlocal so bad
+        t.start_line = sum(line_counts)
+        line_counts.append(t.get_line_count())
+        t.end_line = sum(line_counts)
+        tokens.append(t)
+    for cnc in com_nocom:
+        if not cnc:
+            continue
+        if cnc.startswith('{!') and cnc.endswith('!}'):
+            _add_token(CommentToken(cnc[2:-2]))
+            continue
+        prev_end = 0
+        start = None
+        end = None
+        for match in node_re.finditer(cnc):
+            start, end = match.start(1), match.end(1)
+            if prev_end < start:
+                _add_token(BufferToken(cnc[prev_end:start]))
+            prev_end = end
+            try:
+                _add_token(get_tag(match, inline))
+            except ParseError as pe:
+                pe.line_no = sum(line_counts)
+                raise
+        tail = cnc[prev_end:]
+        if tail:
+            _add_token(BufferToken(tail))
     return tokens
 
-
 #########
 # PARSING
 #########
 
+
 class Section(object):
     def __init__(self, start_tag=None, blocks=None):
         if start_tag is None:
+            refpath = None
             name = '<root>'
         else:
-            name = start_tag.refpath
+            refpath = start_tag.refpath
+            name = start_tag.name
+
+        self.refpath = refpath
         self.name = name
         self.start_tag = start_tag
         self.blocks = blocks or []
@@ -326,7 +361,8 @@ class Section(object):
 
     def to_dust_ast(self):
         symbol = self.start_tag.symbol
-        key = self.start_tag.refpath
+
+        pork = get_path_or_key(self.refpath)
 
         context = ['context']
         contpath = self.start_tag.contpath
@@ -336,7 +372,11 @@ class Section(object):
         params = ['params']
         param_list = self.start_tag.param_list
         if param_list:
-            params.extend(wrap_params(param_list))
+            try:
+                params.extend(params_to_dust_ast(param_list))
+            except ParseError as pe:
+                pe.token = self
+                raise
 
         bodies = ['bodies']
         if self.blocks:
@@ -344,7 +384,7 @@ class Section(object):
                 bodies.extend(b.to_dust_ast())
 
         return [[symbol,
-                [u'key', key],
+                pork,
                 context,
                 params,
                 bodies]]
@@ -408,13 +448,17 @@ class ParseTree(object):
                     ss.append(new_s)
             elif type(token) == ClosingTag:
                 if len(ss) <= 1:
-                    raise ValueError('closing tag before opening tag: %r' % token)
-                if token.refpath != ss[-1].start_tag.refpath:
-                    raise ValueError('nesting error')
+                    msg = 'closing tag before opening tag: %r' % token
+                    raise ParseError(msg, token=token)
+                if token.name != ss[-1].name:
+                    msg = ('improperly nested tags: %r does not close %r' %
+                           (token, ss[-1].start_tag))
+                    raise ParseError(msg, token=token)
                 ss.pop()
             elif type(token) == BlockTag:
                 if len(ss) <= 1:
-                    raise ValueError('cannot start blocks outside of a section')
+                    msg = 'start block outside of a section: %r' % token
+                    raise ParseError(msg, token=token)
                 new_b = Block(name=token.refpath)
                 ss[-1].add(new_b)
             else:
@@ -449,6 +493,13 @@ for nsym in ('#', '?', '^', '<', '+', '@', '%', 'reference',
              'partial', 'context', 'params', 'bodies', 'param'):
     DEFAULT_OPTIMIZERS[nsym] = 'visit'
 
+UNOPT_OPTIMIZERS = dict(DEFAULT_OPTIMIZERS)
+UNOPT_OPTIMIZERS.update({'format': 'noop', 'body': 'visit'})
+
+
+def escape(text, esc_func=json.dumps):
+    return esc_func(text)
+
 
 class Optimizer(object):
     def __init__(self, optimizers=None, special_chars=None):
@@ -505,10 +556,6 @@ class Optimizer(object):
         return self.optimize(node)
 
 
-def escape(text, esc_func=json.dumps):
-    return esc_func(text)
-
-
 #########
 # Compile
 #########
@@ -520,7 +567,28 @@ ROOT_RENDER_TMPL = \
     return {root_func_name}(chk, ctx)
 '''
 
+
+def _python_compile(source, name, global_env=None):
+    if global_env is None:
+        global_env = {}
+    else:
+        global_env = dict(global_env)
+    try:
+        code = compile(source, '<string>', 'single')
+    except:
+        raise
+    if PY3:
+        exec(code, global_env)
+    else:
+        exec("exec code in global_env")
+    return global_env[name]
+
+
 class Compiler(object):
+    """
+    Note: Compiler objects aren't really meant to be reused,
+    the class is just for namespacing and convenience.
+    """
     sections = {'#': 'section',
                 '?': 'exists',
                 '^': 'notexists'}
@@ -531,7 +599,7 @@ class Compiler(object):
 
     def __init__(self, env=None):
         if env is None:
-            env = AshesEnv()
+            env = default_env
         self.env = env
 
         self.bodies = {}
@@ -540,7 +608,11 @@ class Compiler(object):
         self.index = 0
         self.auto = 'h'  # TODO
 
-    def compile(self, ast):  # ast to init?
+    def compile(self, ast, name='render'):
+        python_source = self._gen_python(ast)
+        return _python_compile(python_source, name)
+
+    def _gen_python(self, ast):  # ast to init?
         lines = []
         c_node = self._node(ast)
 
@@ -554,6 +626,7 @@ class Compiler(object):
 
         ret = ROOT_RENDER_TMPL.format(body=body,
                                       root_func_name=c_node)
+        self.python_source = ret
         return ret
 
     def _root_blocks(self):
@@ -568,7 +641,8 @@ class Compiler(object):
         max_body = max(self.bodies.keys())
         ret = [''] * (max_body + 1)
         for i, body in self.bodies.items():
-            ret[i] = '\ndef body_%s(chk, ctx):\n    %sreturn chk%s\n' % (i, self.block_str, body)
+            ret[i] = ('\ndef body_%s(chk, ctx):\n    %sreturn chk%s\n'
+                      % (i, self.block_str, body))
         return ''.join(ret)
 
     def _convert_special(self, node):
@@ -641,7 +715,26 @@ class Compiler(object):
                                          self._node(node[3]))
 
     def _pragma(self, node):
-        raise NotImplemented
+        pr_name = node[1][1]
+        pragma = self.env.pragmas.get(pr_name)
+        if not pragma or not callable(pragma):
+            return ''  # TODO: raise?
+        raw_bodies = node[4]
+        bodies = {}
+        for rb in raw_bodies[1:]:
+            bodies[rb[1][1]] = rb[2]
+
+        raw_params = node[3]
+        params = {}
+        for rp in raw_params[1:]:
+            params[rp[1][1]] = rp[2][1]
+
+        try:
+            ctx = node[2][1][1]
+        except (IndexError, AttributeError):
+            ctx = None
+
+        return pragma(self, ctx, bodies, params)
 
     def _partial(self, node):
         if node[0] == 'body':
@@ -688,20 +781,6 @@ class Compiler(object):
         return escape(node[1])
 
 
-def _python_compile(source, name, global_env=None):
-    if global_env is None:
-        global_env = {}
-    else:
-        global_env = dict(global_env)
-    try:
-        code = compile(source, '<string>', 'single')
-    except:
-        print source
-        raise
-    exec code in global_env
-    return global_env[name]
-
-
 #########
 # Runtime
 #########
@@ -740,14 +819,24 @@ def escape_uri_component(text):
 def sep_helper(chunk, context, bodies):
     if context.stack.index == context.stack.of - 1:
         return chunk
-    return bodies.block(chunk, context)
+    if 'block' in bodies:
+        return bodies['block'](chunk, context)
+    return chunk
 
 
 def idx_helper(chunk, context, bodies):
-    return bodies.block(chunk, context.push(context.stack.index))
+    if 'block' in bodies:
+        return bodies['block'](chunk, context.push(context.stack.index))
+    return chunk
 
 
-DEFAULT_HELPERS = {'sep': sep_helper, 'idx': idx_helper}
+def idx_1_helper(chunk, context, bodies):
+    if 'block' in bodies:
+        return bodies['block'](chunk, context.push(context.stack.index + 1))
+    return chunk
+
+
+DEFAULT_HELPERS = {'sep': sep_helper, 'idx': idx_helper, 'idx_1': idx_1_helper}
 
 
 class Context(object):
@@ -839,11 +928,11 @@ class Context(object):
 
 
 class Stack(object):
-    def __init__(self, head, tail=None, index=0, length=1):
+    def __init__(self, head, tail=None, index=None, length=None):
         self.head = head
         self.tail = tail
-        self.index = index
-        self.of = length
+        self.index = index or 0
+        self.of = length or 1
         #self.is_object = is_scalar(head)
 
     def __repr__(self):
@@ -962,8 +1051,7 @@ class Chunk(object):
     def reference(self, elem, context, auto, filters=None):
         if callable(elem):
             # this whole callable thing is a pretty big TODO
-            elem = elem(self, context, None, {'auto': auto,
-                                              'filters': filters})
+            elem = elem(self, context)
             if isinstance(elem, Chunk):
                 return elem
         if is_empty(elem):
@@ -986,7 +1074,7 @@ class Chunk(object):
             # on sections referencing empty lists.
             return else_body(self, context)
 
-        if not body:
+        if not body or elem is None:
             return self
         if is_scalar(elem) or hasattr(elem, 'keys'):  # haaack
             if elem is True:
@@ -1031,7 +1119,7 @@ class Chunk(object):
         return context.env.load_chunk(elem, self, context)
 
     def helper(self, name, context, bodies, params=None):
-        return context.env.helpers[name](self, context, bodies, params)
+        return context.env.helpers[name](self, context, bodies)
 
     def capture(self, body, context, callback):
         def map_func(chunk):
@@ -1073,16 +1161,172 @@ DEFAULT_FILTERS = {
     'uc': escape_uri_component}
 
 
+#########
+# Pragmas
+#########
+
+
+def esc_pragma(compiler, context, bodies, params):
+    old_auto = compiler.auto
+    if not context:
+        context = 'h'
+    if context == 's':
+        compiler.auto = ''
+    else:
+        compiler.auto = context
+    out = compiler._parts(bodies['block'])
+    compiler.auto = old_auto
+    return out
+
+
+DEFAULT_PRAGMAS = {
+    'esc': esc_pragma
+}
+
+
 ###########
 # Interface
 ###########
-class AshesEnv(object):
+
+
+class Template(object):
     def __init__(self,
-                 filters=None,
+                 name,
+                 source,
+                 source_file=None,
+                 optimize=True,
+                 keep_source=True,
+                 env=None,
+                 lazy=False):
+        self.name = name
+        self.source = source
+        self.source_file = source_file
+        self.last_mtime = None
+        if source_file:
+            self.last_mtime = os.path.getmtime(source_file)
+        self.optimized = optimize
+        if env is None:
+            env = default_env
+        self.env = env
+
+        if lazy:  # lazy is really only for testing
+            self.render_func = None
+            return
+        self.render_func = self._get_render_func(optimize)
+        if not keep_source:
+            self.source = None
+
+    def render(self, model, env=None):
+        env = env or self.env
+        rendered = []
+
+        def tmp_cb(err, result):
+            if err:
+                print('Error on template %r: %r' % (self.name, err))
+                raise RenderException(err)
+            else:
+                rendered.append(result)
+                return result
+
+        chunk = Stub(tmp_cb).head
+        self.render_chunk(chunk, Context.wrap(env, model)).end()
+        return rendered[0]
+
+    def render_chunk(self, chunk, context):
+        if not self.render_func:
+            # to support laziness for testing
+            self.render_func = self._get_render_func()
+        return self.render_func(chunk, context)
+
+    def _get_tokens(self):
+        if not self.source:
+            return None
+        return tokenize(self.source)
+
+    def _get_ast(self, optimize=False, raw=False):
+        if not self.source:
+            return None
+        try:
+            dast = ParseTree.from_source(self.source).to_dust_ast()
+        except ParseError as pe:
+            pe.source_file = self.source_file
+            raise
+        if raw:
+            return dast
+        return self.env.filter_ast(dast, optimize)
+
+    def _get_render_func(self, optimize=True, ret_str=False):
+        # switching over to optimize=True by default because it
+        # makes the output easier to read and more like dust's docs
+        ast = self._get_ast(optimize)
+        if not ast:
+            return None
+        compiler = Compiler(self.env)
+        func = compiler.compile(ast)
+        if ret_str:
+            # for testing/dev purposes
+            return Compiler(self.env)._gen_python(ast)
+        return func
+
+
+class AshesException(Exception):
+    pass
+
+
+class TemplateNotFound(AshesException):
+    pass
+
+
+class RenderException(AshesException):
+    pass
+
+
+class ParseError(AshesException):
+    token = None
+    source_file = None
+
+    def __init__(self, message, line_no=None, token=None):
+        self.message = message
+        self.token = token
+        self._line_no = line_no
+
+        super(ParseError, self).__init__(self.__str__())
+
+    @property
+    def line_no(self):
+        if self._line_no:
+            return self._line_no
+        if getattr(self.token, 'start_line', None) is not None:
+            return self.token.start_line
+        return None
+
+    @line_no.setter
+    def set_line_no(self, val):
+        self._line_no = val
+
+    def __str__(self):
+        msg = self.message
+        infos = []
+        if self.source_file:
+            infos.append('in %s' % self.source_file)
+        if self.line_no is not None:
+            infos.append('line %s' % self.line_no)
+        if infos:
+            msg += ' (%s)' % ' - '.join(infos)
+        return msg
+
+
+class BaseAshesEnv(object):
+    def __init__(self,
+                 loaders=None,
                  helpers=None,
+                 filters=None,
                  special_chars=None,
-                 optimizers=None):
+                 optimizers=None,
+                 pragmas=None,
+                 auto_reload=True):
         self.templates = {}
+        self.loaders = list(loaders or [])
         self.filters = dict(DEFAULT_FILTERS)
         if filters:
             self.filters.update(filters)
@@ -1095,19 +1339,71 @@ class AshesEnv(object):
         self.optimizers = dict(DEFAULT_OPTIMIZERS)
         if optimizers:
             self.optimizers.update(optimizers)
+        self.pragmas = dict(DEFAULT_PRAGMAS)
+        if pragmas:
+            self.pragmas.update(pragmas)
+        self.auto_reload = auto_reload
 
-    def compile(self, source, name, src_file=None):
-        comp_str = self._compile_str(source)
-        tmpl_func = _python_compile(comp_str, 'render')
-        self.templates[name] = tmpl_func
-        return comp_str
+    def render(self, name, model):
+        tmpl = self.load(name)
+        return tmpl.render(model, self)
 
-    def _compile_str(self, source):
-        parse_tree = ParseTree.from_source(source)
-        dust_ast = parse_tree.to_dust_ast()
-        optimized_ast = Optimizer().optimize(dust_ast)
-        comp_str = Compiler().compile(optimized_ast)
-        return comp_str
+    def load(self, name):
+        try:
+            template = self.templates[name]
+        except KeyError:
+            template = self._load_template(name)
+            self.register(template)
+        if self.auto_reload:
+            if not getattr(template, 'source_file', None):
+                return template
+            mtime = os.path.getmtime(template.source_file)
+            if mtime > template.last_mtime:
+                template = self._load_template(name)
+                self.register(template)
+        return self.templates[name]
+
+    def _load_template(self, name):
+        for loader in self.loaders:
+            try:
+                source = loader.load(name)
+            except TemplateNotFound:
+                continue
+            else:
+                return source
+        raise TemplateNotFound(name)
+
+    def load_all(self, do_register=True, **kw):
+        all_tmpls = []
+        for loader in reversed(self.loaders):
+            # reversed so the first loader to have a template
+            # will take precendence on registration
+            if callable(getattr(loader, 'load_all', None)):
+                tmpls = loader.load_all(self, **kw)
+                all_tmpls.extend(tmpls)
+                if do_register:
+                    for t in tmpls:
+                        self.register(t)
+        return all_tmpls
+
+    def register(self, template, name=None):
+        if name is None:
+            name = template.name
+        self.templates[name] = template
+        return
+
+    def register_source(self, name, tmpl_src):
+        tmpl = Template(name, tmpl_src, env=self)
+        self.register(tmpl)
+        return tmpl
+
+    def filter_ast(self, ast, optimize=True):
+        if optimize:
+            optimizers = self.optimizers
+        else:
+            optimizers = UNOPT_OPTIMIZERS
+        optimizer = Optimizer(optimizers, self.special_chars)
+        return optimizer.optimize(ast)
 
     def apply_filters(self, string, auto, filters):
         filters = filters or []
@@ -1120,29 +1416,123 @@ class AshesEnv(object):
         return string
 
     def load_chunk(self, name, chunk, context):
-        if name not in self.templates:
-            # TODO: on_load hook
-            return chunk.set_error(Exception('Template not found "%s"' % name))
-        return self.templates[name](chunk, context)
-
-
-    def render(self, name, model):
         try:
-            template = self.templates[name]
-        except KeyError:
-            raise ValueError('No template named "%s"' % name)
-        rendered = []
+            tmpl = self.load(name)
+        except TemplateNotFound as tnf:
+            return chunk.set_error(tnf)
+        return tmpl.render_chunk(chunk, context)
 
-        def tmp_cb(err, result):
-            if err:
-                print 'Error on template %r: %r' % (name, err)
-                raise Exception(err)
-            else:
-                rendered.append(result)
-                return result
 
-        chunk = Stub(tmp_cb).head
-        template(chunk, Context.wrap(self, model)).end()
-        return rendered[0]
+class AshesEnv(BaseAshesEnv):
+    """
+    A slightly more accessible Ashes environment, with more
+    user-friendly options exposed.
+    """
+    def __init__(self, paths=None, keep_whitespace=False, *a, **kw):
+        self.paths = list(paths or [])
+        self.keep_whitespace = keep_whitespace
+        exts = list(kw.pop('exts', DEFAULT_EXTENSIONS))
+
+        super(AshesEnv, self).__init__(*a, **kw)
+
+        for path in self.paths:
+            tpl = TemplatePathLoader(path, exts)
+            self.loaders.append(tpl)
+
+    def filter_ast(self, ast, optimize=None):
+        optimize = not self.keep_whitespace  # preferences override
+        return super(AshesEnv, self).filter_ast(ast, optimize)
+
+    def load_all(self):
+        ret = []
+        for loader in self.loaders:
+            ret.extend(loader.load_all())
+        return ret
+
+
+DEFAULT_EXTENSIONS = ('.dust', '.html', '.xml')
+
+
+def walk_ext_matches(path, exts=None):
+    if exts is None:
+        exts = DEFAULT_EXTENSIONS
+    exts = list(['.' + e.lstrip('.') for e in exts])
+    matches = []
+    for root, _, filenames in os.walk(path):
+        for fn in filenames:
+            for ext in exts:
+                if fn.endswith(ext):
+                    matches.append(os.path.join(root, fn))
+    return matches
+
+
+class TemplatePathLoader(object):
+    def __init__(self, root_path, exts=None, encoding='utf-8'):
+        self.root_path = os.path.normpath(root_path)
+        self.encoding = encoding
+        self.exts = exts or dict(DEFAULT_EXTENSIONS)
+
+    def load(self, path, env=None):
+        env = env or default_env
+        norm_path = os.path.normpath(path)
+        if path.startswith('../'):
+            raise ValueError('no traversal above loader root path: %r' % path)
+        if not path.startswith(self.root_path):
+            norm_path = os.path.join(self.root_path, norm_path)
+        abs_path = os.path.abspath(norm_path)
+        if os.path.isfile(abs_path):
+            with codecs.open(abs_path, 'r', self.encoding) as f:
+                source = f.read()
+        else:
+            raise TemplateNotFound(path)
+        template = Template(path, source, abs_path, env=env)
+        return template
+
+    def load_all(self, env, exts=None, **kw):
+        ret = []
+        exts = exts or self.exts
+        tmpl_paths = walk_ext_matches(self.root_path, exts)
+        for tmpl_path in tmpl_paths:
+            ret.append(self.load(tmpl_path, env))
+        return ret
+
+
+class FlatteningPathLoader(TemplatePathLoader):
+    """
+    I've seen this mode of using dust templates in a couple places,
+    but really it's lazy and too ambiguous. It increases the chances
+    of silent conflicts and makes it hard to tell which templates refer
+    to which just by looking at the template code.
+    """
+    def __init__(self, *a, **kw):
+        self.keep_ext = kw.pop('keep_ext', True)
+        super(FlatteningPathLoader, self).__init__(*a, **kw)
+
+    def load(self, *a, **kw):
+        tmpl = super(FlatteningPathLoader, self).load(*a, **kw)
+        name = os.path.basename(tmpl.name)
+        if not self.keep_ext:
+            name, ext = os.path.splitext(name)
+        tmpl.name = name
+        return tmpl
+
 
 ashes = default_env = AshesEnv()
+
+
+def _main():
+    try:
+        tpl = FlatteningPathLoader('./_test_tmpls', keep_ext=False)
+        ashes.loaders.append(tpl)
+        ashes.load_all()
+
+        rendereds = dict([(k, t.render({})) for k, t in ashes.templates.items()])
+    except Exception as e:
+        import pdb;pdb.post_mortem()
+        raise
+    else:
+        pass
+        #import pdb;pdb.set_trace()
+
+if __name__ == '__main__':
+    _main()
